#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import copy
import enum
from abc import ABC, abstractmethod
from collections.abc import ItemsView, Iterator, KeysView, Mapping, MutableMapping, Sequence, ValuesView
from dataclasses import dataclass, field
from datetime import datetime
from typing import Protocol, TypeVar, overload
from uuid import UUID

from .exceptions import ServiceHealthCheckFailed

__all__ = [
    "DependencyStatus",
    "EmptyResponse",
    "Environment",
    "HTTPHeaderDict",
    "HTTPResponse",
    "HealthCheckType",
    "Page",
    "RequestMethod",
    "ResourceMetadata",
    "ServiceHealth",
    "ServiceStatus",
    "ServiceUser",
]


class RequestMethod(str, enum.Enum):
    """HTTP request method."""

    GET = "GET"
    """HTTP [`GET`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/GET)"""

    HEAD = "HEAD"
    """HTTP [`HEAD`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/HEAD)"""

    POST = "POST"
    """HTTP [`POST`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/POST)"""

    PUT = "PUT"
    """HTTP [`PUT`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PUT)"""

    DELETE = "DELETE"
    """HTTP [`DELETE`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/DELETE)"""

    OPTIONS = "OPTIONS"
    """HTTP [`OPTIONS`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/OPTIONS)"""

    PATCH = "PATCH"
    """HTTP [`PATCH`](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods/PATCH)"""

    def __str__(self) -> str:
        return self.value


class HTTPHeaderDict(MutableMapping[str, str]):
    def __init__(self, seq: Mapping[str, str] | Sequence[tuple[str, str]] | None = None, **kwargs: str) -> None:
        self.__values: dict[str, str] = {}
        self.update(seq, **kwargs)

    def update(self, seq: Mapping[str, str] | Sequence[tuple[str, str]] | None = None, **kwargs: str) -> None:
        if isinstance(seq, Mapping):
            self.__update_from_mapping(seq)
        elif isinstance(seq, Sequence):
            self.__update_from_sequence(seq)

        self.__update_from_mapping(kwargs)

    def __update_from_mapping(self, mapping: Mapping[str, str]) -> None:
        for key, value in mapping.items():
            self.__setitem__(key, value)

    def __update_from_sequence(self, seq: Sequence[tuple[str, str]]) -> None:
        for key, value in seq:
            self.__setitem__(key, value)

    def __setitem__(self, key: str, value: str) -> None:
        lookup = key.title()
        if lookup in self.__values and lookup != "Set-Cookie":
            # RFC 7230 section 3.2.2: Field Order.
            # A recipient MAY combine multiple header fields with the same field
            # name into one "field-name: field-value" pair, without changing the
            # semantics of the message, by appending each subsequent field value to
            # the combined field value in order, separated by a comma.  The order
            # in which header fields with the same field name are received is
            # therefore significant to the interpretation of the combined field
            # value;
            # ...
            # In practice, the "Set-Cookie" header field ([RFC6265]) often
            # appears multiple times in a response message and does not use the
            # list syntax, violating the above requirements on multiple header
            # fields with the same name.  Since it cannot be combined into a
            # single field-value, recipients ought to handle "Set-Cookie" as a
            # special case while processing header fields.
            # https://www.rfc-editor.org/rfc/rfc7230#section-3.2.2
            self.__values[lookup] += "," + value
        else:
            self.__values[lookup] = value

    def __delitem__(self, key: str) -> None:
        del self.__values[key.title()]

    def __getitem__(self, key: str) -> str:
        return self.__values[key.title()]

    def __len__(self) -> int:
        return len(self.__values)

    def __iter__(self) -> Iterator[str]:
        return iter(self.keys())

    def __contains__(self, item: str) -> bool:
        return item.title() in self.__values

    def __repr__(self) -> str:
        repr_data = {}
        for key, value in self.items():
            if key in ("Authorization", "Proxy-Authorization", "Cookie", "Set-Cookie"):
                # Do not expose sensitive information.
                value = "*****"
            repr_data[key] = value

        return f"{self.__class__.__name__}({repr_data!r})"

    def items(self) -> ItemsView[str, str]:
        return ItemsView(self)

    def keys(self) -> KeysView[str]:
        return KeysView(self.__values)

    def values(self) -> ValuesView[str]:
        return ValuesView(self.__values)

    def copy(self) -> HTTPHeaderDict:
        return copy.deepcopy(self)


@dataclass(frozen=True, kw_only=True)
class EmptyResponse:
    status: int
    reason: str | None = None
    headers: HTTPHeaderDict = field(default_factory=HTTPHeaderDict)

    def getheaders(self) -> HTTPHeaderDict:
        return self.headers.copy()

    def getheader(self, key: str, default: str | None = None) -> str:
        return self.headers.get(key, default)


@dataclass(frozen=True, kw_only=True)
class HTTPResponse(EmptyResponse):
    data: bytes


@dataclass(frozen=True, kw_only=True)
class Environment:
    """Service environment configuration.

    This class is used by service clients to determine the organisation and workspace context for requests.
    """

    hub_url: str
    """The base URL of the Evo service hub."""

    org_id: UUID
    """The organisation ID for the current environment."""

    workspace_id: UUID
    """The workspace ID for the current environment."""


class _UserModel(Protocol):
    id: UUID
    name: str | None
    email: str | None


@dataclass(frozen=True, kw_only=True)
class ServiceUser:
    """Metadata about a user."""

    id: UUID
    """The user UUID."""

    name: str | None
    """The user display name."""

    email: str | None
    """The user email address."""

    @classmethod
    def from_model(cls, model: _UserModel) -> ServiceUser:
        """Create a new instance from an instance of a generated model."""
        return cls(id=model.id, name=model.name, email=model.email)


@dataclass(frozen=True, kw_only=True)
class ResourceMetadata(ABC):
    """Metadata about a resource stored in an Evo service."""

    environment: Environment
    """The environment in which the resource is stored."""

    id: UUID
    """The resource UUID."""

    name: str
    """The resource name."""

    created_at: datetime
    """The resource creation timestamp."""

    created_by: ServiceUser | None = None
    """The user who created the resource."""

    @property
    @abstractmethod
    def url(self) -> str:
        """The URL of the resource.

        Resource URLs should point to the appropriate API endpoint to get the resource metadata.
        Prefer URLs that reference resources by UUID and include the resource version.
        """
        pass


_Metadata = TypeVar("_Metadata", bound=ResourceMetadata)


class Page(Sequence[_Metadata]):
    """A page of resource metadata from a paginated response.

    This type exposes paginated metadata from an API response, including the offset, limit, and total number of items.
    The contained items are the actual metadata from the response.
    """

    def __init__(self, *, offset: int, limit: int, total: int, items: Sequence[_Metadata]) -> None:
        self._offset = offset
        self._limit = limit
        self._total = total
        self._items = items

    @property
    def offset(self) -> int:
        """The offset of the first item in the page."""
        return self._offset

    @property
    def limit(self) -> int:
        """The maximum number of items per page."""
        return self._limit

    @property
    def size(self) -> int:
        """The number of items in the page."""
        return len(self._items)

    def __len__(self) -> int:
        """The number of items in the page."""
        return self.size

    @property
    def total(self) -> int:
        """The total number of items that are available, as reported by the API."""
        return self._total

    def items(self) -> list[_Metadata]:
        """Get the items that are in the page.

        Items are copied to prevent modification of the original items.

        :returns: A list of items in the page.
        """
        return [copy.deepcopy(item) for item in self._items]

    @overload
    def __getitem__(self, key: int) -> _Metadata: ...

    @overload
    def __getitem__(self, key: slice) -> list[_Metadata]: ...

    def __getitem__(self, key: int | slice) -> _Metadata | list[_Metadata]:
        """Get an item or items from the page.

        If the key is an integer, the item at that index is returned. If the key is a slice, a list of items in the
        slice are returned.

        Items are copied to prevent modification of the original items.

        :param key: The index of the item to get, or a slice of items to get.

        :returns: The item or items from the page.
        """
        if isinstance(key, int):
            return copy.deepcopy(self._items[key])
        elif isinstance(key, slice):
            return [copy.deepcopy(item) for item in self._items[key]]
        else:
            raise TypeError(f"Invalid key type: {type(key)}")

    @property
    def next_offset(self) -> int:
        """The offset of the next page."""
        return self.offset + self.size

    @property
    def is_last(self) -> bool:
        """Whether the page is the last page."""
        return self.next_offset >= self.total


class HealthCheckType(enum.Enum):
    """The type of health check to be performed.

    This enum determines which parameters are provided in a service health check request, which in turn determines the
    type of health check that is performed by the service.
    """

    BASIC = enum.auto()
    """Perform a basic health check, ignoring service dependencies."""

    FULL = enum.auto()
    """Perform a full health check, including service dependencies.
    A failed dependency may not necessarily result in a failed health check. If a dependency is unhealthy but the
    service can still handle requests, the service may report report a degraded status.
    """

    STRICT = enum.auto()
    """Perform a strict health check, including service dependencies.
    A strict health check will always fail if any dependency is unhealthy.
    """


class ServiceStatus(str, enum.Enum):
    """Service status enumeration."""

    HEALTHY = "pass"
    """The service is healthy and able to handle requests."""

    DEGRADED = "degraded"
    """The service is running but some features may not work, or performance may be reduced."""

    UNHEALTHY = "fail"
    """The service is unable to handle requests."""


class DependencyStatus(str, enum.Enum):
    """Dependency status enumeration."""

    HEALTHY = "pass"
    """The dependency is healthy and able to handle requests."""

    UNHEALTHY = "fail"
    """The dependency is unable to handle requests."""


@dataclass(frozen=True, kw_only=True)
class ServiceHealth:
    service: str
    """The name of the service."""

    status_code: int
    """The status code of the health check response."""

    status: ServiceStatus
    """The current status of the service."""

    version: str
    """An arbitrary string representing the service version."""

    dependencies: dict[str, DependencyStatus] | None
    """A dictionary of service dependencies and their status."""

    def raise_for_status(self) -> None:
        """Raise an exception if the service is unhealthy.

        :raises ServiceHealthCheckFailed: If `status_code` is not 200.
        """
        if self.status_code != 200:
            msg = f"Health check for {self.service} service failed with status {self.status} ({self.status_code})"
            if self.dependencies is not None:
                for dependency, status in self.dependencies.items():
                    msg += f"\n  {dependency}: {status}"
            raise ServiceHealthCheckFailed(msg)


class OrderByOperatorEnum(str, enum.Enum):
    """Enumeration for order_by operators for listing endpoints."""

    asc = "asc"
    desc = "desc"



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .cache import Cache
from .data import parse_order_by
from .feedback import NoFeedback, PartialFeedback, iter_with_fb
from .health_check import get_service_health, get_service_status
from .retry import BackoffExponential, BackoffIncremental, BackoffLinear, BackoffMethod, Retry, RetryHandler
from .version import get_header_metadata

__all__ = [
    "BackoffExponential",
    "BackoffIncremental",
    "BackoffLinear",
    "BackoffMethod",
    "Cache",
    "NoFeedback",
    "PartialFeedback",
    "Retry",
    "RetryHandler",
    "get_header_metadata",
    "get_service_health",
    "get_service_status",
    "iter_with_fb",
    "parse_order_by",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..exceptions import EvoClientException

__all__ = [
    "ChunkedIOError",
    "ChunkedIOException",
    "RenewalError",
    "RenewalTimeoutError",
]


class ChunkedIOException(EvoClientException):
    """Base exception for unrecoverable errors that are expected within io.

    These exceptions should be caught by client code.
    """


class ChunkedIOError(ChunkedIOException):
    """Base exception for recoverable errors that are handled within chunked_io.

    All ChunkedIOError subtypes are considered recoverable and should implement a recover() method that prepares the
    `evo.common.io.interfaces.ISource` or `evo.common.io.interfaces.IDestination` that raised
    the error for a retry attempt.

    :ivar raised_by: The source or destination that raised the exception.
    """

    def __init__(self, message: str) -> None:
        """
        :param message: The error message.
        """
        super().__init__(message)

    async def recover(self) -> bool:
        """Run the steps required to recover from the original error and prepare for a retry attempt."""
        return False  # ChunkedIOError is not recoverable by default.


class RenewalError(EvoClientException):
    """RenewalError is raised when renewing the url fails."""


class RenewalTimeoutError(RenewalError):
    """RenewalTimeoutError is raised when an attempt to renew a url is made within a specified threshold."""


class DataExistsError(EvoClientException):
    """Raised when requesting an upload url for a binary resource that already exists in the target service."""


class DataNotFoundError(EvoClientException):
    """Raised when requesting a download url for a binary resource that does not exist in the target service."""



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import unittest
from datetime import datetime, time, timezone
from typing import TypeVar


def utc_datetime(
    year: int, month: int = 1, day: int = 1, hour: int = 0, minute: int = 0, second: int = 0, microsecond: int = 0
) -> datetime:
    """Create a datetime object in UTC time."""
    return datetime(year, month, day, hour, minute, second, microsecond, tzinfo=timezone.utc)


def utc_time(hour: int = 0, minute: int = 0, second: int = 0, microsecond: int = 0) -> time:
    """Create a time object in UTC time."""
    return time(hour, minute, second, microsecond, tzinfo=timezone.utc)


_FT = TypeVar("_FT")


def long_test(test: _FT) -> _FT:
    """Decorator to mark a test as long-running.

    This decorator should be used to mark tests that take a long time to run, or that require a lot of resources. The
    decorator will cause the test to be skipped by default, and will be run only if the `CI` environment variable is
    set (as is the case in GitHub workflows).
    """
    return unittest.skipUnless("CI" in os.environ, "Make local testing faster")(test)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json

from ..connector import APIConnector
from ..data import DependencyStatus, HealthCheckType, HTTPResponse, RequestMethod, ServiceHealth, ServiceStatus
from ..exceptions import ClientValueError

__all__ = [
    "get_service_health",
    "get_service_status",
]


async def _check_service_health(
    connector: APIConnector, service_name: str, full: bool, check_type: HealthCheckType
) -> HTTPResponse:
    """Internal that calls the health check endpoint of a named API.

    :param connector: The API connector.
    :param service_name: The name of the service to check, as used in the Evo service URL.
    :param full: Whether to include full health check information.
    :param check_type: The type of health check to perform.

    :return: The response from the health check endpoint.
    """
    query_params = {}
    if full is True:
        query_params["full"] = "true"
    if check_type in {HealthCheckType.FULL, HealthCheckType.STRICT}:
        query_params["check_dependencies"] = "true"
    if check_type is HealthCheckType.STRICT:
        query_params["strict"] = "true"

    return await connector.call_api(
        RequestMethod.GET,
        f"/{service_name}/health_check",
        query_params=query_params,
        response_types_map={"200": HTTPResponse, "503": HTTPResponse},
    )


def _parse_service_status(status: str) -> ServiceStatus:
    """Parse a service status string into a ServiceStatus object.

    :param status: The status string to parse.

    :return: A ServiceStatus object.
    """
    try:
        return ServiceStatus(status)
    except ValueError as err:
        raise ClientValueError(f"Invalid service status: {status!r}") from err


def _parse_service_health_dict(service_name: str, status_code: int, response: dict) -> ServiceHealth:
    """Parse a service health response into a ServiceHealth object.

    :param service_name: The name of the service that was checked.
    :param status_code: The status code of the health check response.
    :param response: The response from the health check endpoint.

    :return: A ServiceHealth object.
    """
    try:
        status_str = response["status"]
        version = response["version"]
    except KeyError as err:
        raise ClientValueError("Service health response must include a status.") from err

    service_status = _parse_service_status(status_str)

    dependencies = None
    if "dependencies" in response:
        dependencies = {}
        for name, status_str in response["dependencies"].items():
            try:
                dependency_status = DependencyStatus(status_str)
            except ValueError as err:
                raise ClientValueError(f"Invalid status for dependency '{name}': {status_str!r}") from err
            dependencies[name] = dependency_status

    return ServiceHealth(
        service=service_name, status_code=status_code, status=service_status, version=version, dependencies=dependencies
    )


async def get_service_health(
    connector: APIConnector, service_name: str, check_type: HealthCheckType = HealthCheckType.FULL
) -> ServiceHealth:
    """Check the health of a named service.

    Performs a full service health check and returns the results.

    :param connector: The API connector.
    :param service_name: The name of the service to check, as used in the Evo service URL.
    :param check_type: The type of health check to perform.

    :return: A ServiceHealth object.

    :raises EvoAPIException: If the API returns an unexpected status code.
    :raises ClientValueError: If the response is not a valid service health check response.
    """
    response = await _check_service_health(connector, service_name, full=True, check_type=check_type)
    try:
        response_data = json.loads(response.data)
    except ValueError as err:
        raise ClientValueError("Invalid service health check response.") from err
    else:
        return _parse_service_health_dict(service_name, response.status, response_data)


async def get_service_status(
    connector: APIConnector, service_name: str, check_type: HealthCheckType = HealthCheckType.FULL
) -> ServiceStatus:
    """Get the status of a named service.

    Performs a basic service health check and returns the status.

    :param connector: The API connector.
    :param service_name: The name of the service to check, as used in the Evo service URL.
    :param check_type: The type of health check to perform.

    :return: The status of the service.

    :raises EvoAPIException: If the API returns an unexpected status code.
    :raises ClientValueError: If the response is not a valid service health status.
    """
    response = await _check_service_health(connector, service_name, full=False, check_type=check_type)
    try:
        response_str = response.data.decode()
    except ValueError as err:
        raise ClientValueError("Invalid service status response.") from err
    else:
        return _parse_service_status(response_str)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import datetime
import functools
import json
import re
from collections.abc import Mapping
from enum import Enum
from inspect import isclass
from types import GenericAlias, NoneType, TracebackType
from typing import Any, ParamSpec, TypeVar
from urllib.parse import quote, urlencode
from uuid import UUID

from dateutil.parser import parse
from pydantic import BaseModel

from evo import logging

from .data import EmptyResponse, HTTPHeaderDict, HTTPResponse, RequestMethod
from .exceptions import (
    ClientTypeError,
    ClientValueError,
    CustomTypedError,
    EvoAPIException,
    EvoClientException,
    GeneralizedTypedError,
    UnauthorizedException,
    UnknownResponseError,
)
from .interfaces import IAuthorizer, ITransport

logger = logging.getLogger("connector")

__all__ = [
    "APIConnector",
    "NoAuth",
]

T = TypeVar("T")
P = ParamSpec("P")


def retry_on_auth_error(func):  # No type annotation to prevent hiding the signature of the decorated function.
    @functools.wraps(func)
    async def wrapper(self: APIConnector, *args: P.args, **kwargs: P.kwargs) -> T:
        # Always use the connector in a context manager to ensure the transport is opened and closed correctly.
        # ITransport implementations are required to be re-entrant, so this is safe. It may still be useful to open
        # and close the transport elsewhere in client code to prevent unnecessarily destroying and recreating any
        # underlying sessions.
        async with self:
            try:
                return await func(self, *args, **kwargs)
            except UnauthorizedException:
                logger.debug("Unauthorized exception caught. Attempting to refresh the access token", exc_info=True)
                if not await self._authorizer.refresh_token():
                    logger.debug("Failed to refresh the access token.", exc_info=True)
                    raise
                else:
                    logger.debug("Access token refreshed. Retrying the request.")
                    return await func(self, *args, **kwargs)
            except:
                logger.debug("An error occurred while calling the API.", exc_info=True)
                raise

    return wrapper


class _NoAuth(IAuthorizer):
    """An authorizer that does not provide any authentication."""

    async def get_default_headers(self) -> HTTPHeaderDict:
        """Return an empty header dictionary."""
        return HTTPHeaderDict()

    async def refresh_token(self) -> bool:
        """Return False, as there is no token to refresh."""
        return False


NoAuth = _NoAuth()
"""An authorizer that does not provide any authentication."""


class APIConnector:
    """Generic client for facilitating API requests."""

    def __init__(
        self,
        base_url: str,
        transport: ITransport,
        authorizer: IAuthorizer = NoAuth,
        additional_headers: Mapping[str, Any] | None = None,
    ) -> None:
        """
        :param base_url: The host URL of the API.
        :param transport: The transport to use for sending requests.
        :param authorizer: The authorizer to use for authenticating requests.
        :param additional_headers: Additional headers to include in each request.
        """
        self._base_url = base_url.rstrip("/")
        self._transport = transport
        self._authorizer = authorizer
        self._additional_headers = additional_headers

    @property
    def base_url(self) -> str:
        """The base_url of the connected API."""
        return self._base_url + "/"

    @property
    def transport(self) -> ITransport:
        """The transport used to send requests."""
        return self._transport

    async def open(self) -> None:
        """Open the HTTP transport."""
        await self._transport.open()

    async def close(self) -> None:
        """Close the HTTP transport."""
        await self._transport.close()

    async def __aenter__(self) -> APIConnector:
        await self.open()
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> None:
        await self.close()

    @retry_on_auth_error
    async def call_api(
        self,
        method: RequestMethod,
        resource_path: str,
        path_params: Mapping[str, Any] | None = None,
        query_params: Mapping[str, Any] | None = None,
        header_params: Mapping[str, Any] | None = None,
        post_params: Mapping[str, Any] | None = None,
        body: object | str | bytes | None = None,
        collection_formats: Mapping[str, str] | None = None,
        response_types_map: Mapping[str, type[T]] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> T:
        """Call the API with the given parameters and deserialize the response.

        Errors raised by `ITransport.request` are not handled by this method.

        :param method: HTTP request method.
        :param resource_path: Path to the API endpoint.
        :param path_params: Path parameters to embed in the url.
        :param query_params: Query parameters to embed in the url.
        :param header_params: Header parameters to be placed in the request header.
        :param post_params: Post request form parameters.
        :param body: Body to send with the request.
        :param collection_formats: Dict of collection formats for path, query, header, and post parameters, where the
            key is the parameter name, and the value is one of 'multi', 'ssv', 'pipes', or 'csv'.
            'csv' is the default collection format.
        :param response_types_map: Mapping of response status codes to response data types. The response will
            be deserialized to the corresponding type.
        :param request_timeout: Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The deserialized response, in the format determined by the response types map.

        :raise BadRequestException: If the server responds with HTTP status 400.
        :raise UnauthorizedException: If the server responds with HTTP status 401.
        :raise ForbiddenException: If the server responds with HTTP status 403.
        :raise NotFoundException: If the server responds with HTTP status 404.
        :raise BaseTypedError: If the server responds with any other HTTP status between 400 and 599, and the body
            of the response has a type.
        :raise EvoAPIException: If the server responds with any other HTTP status between 400 and 599, and the body of
            the response does not have a type.
        :raise UnknownResponseError: For other HTTP status codes with no corresponding response type in
            `response_types_map`.
        """
        collection_formats = collection_formats or {}

        # Process URL parameters.
        resource_url = self._encode_url_parameters(resource_path, path_params, query_params, collection_formats)

        # Process request headers.
        headers = await self._authorizer.get_default_headers()
        headers.update(self._additional_headers)
        if header_params is not None:
            specified_headers = set()  # So we make sure each header is only overridden once.
            for key, value in self._parameters_to_tuples(header_params, collection_formats):
                if key in specified_headers or isinstance(value, list):
                    raise RuntimeError(f"Multiple values not supported in header '{key}'")
                else:
                    headers[key] = value
                    specified_headers.add(key)

        # Process post parameters.
        post_params = self._parameters_to_tuples(post_params, collection_formats) if post_params else None

        # Sanitize body
        body = self._sanitize_for_serialization(body) if body else None

        # Perform request.
        logger.debug(f"Making {method} request to {resource_url}")
        response = await self._transport.request(
            method=method,
            url=resource_url,
            headers=headers,
            post_params=post_params,
            body=body,
            request_timeout=request_timeout,
        )

        # Prepare response type.
        if (default_response_type := GeneralizedTypedError.from_status_code(response.status)) is not None:
            pass  # Use the response type returned above.
        elif 400 <= response.status <= 599:  # Error status code.
            default_response_type = EvoAPIException
        else:
            default_response_type = UnknownResponseError

        if response_types_map is not None:
            # Always use the type from response_types_map if it is available.
            response_type = response_types_map.get(str(response.status), default_response_type)
        else:
            response_type = default_response_type

        # Decode response.
        response_object = self._deserialize(response, response_type)

        return response_object

    def _encode_url_parameters(
        self,
        resource_path: str,
        path_params: dict[str, Any] | None = None,
        query_params: dict[str, Any] | None = None,
        collection_formats: dict[str, str] | None = None,
    ) -> str:
        """Encode path and query parameters within the resource path.

        :param resource_path: The resource path.
        :param path_params: Parameters that are used to replace template parameters in the resource path.
        :param query_params: Query parameters that are to be encoded at the end of the URL.
        :param collection_formats: Dict of collection formats for path and query parameters.

        :return: A URL with all path and query parameters encoded.
        """
        resource_url = self._base_url + "/" + resource_path.lstrip("/")

        if path_params:
            for key, value in self._parameters_to_tuples(path_params, collection_formats):
                resource_url = resource_url.replace(f"{{{key}}}", quote(str(value)))

        if query_params:
            resource_url += "?" + self._parameters_to_url_query(query_params, collection_formats)

        return resource_url

    @classmethod
    def _sanitize_for_serialization(cls, obj: Any | None) -> Any | None:
        """Builds a JSON object for serialization.

        If obj is None return None.
        If obj is an Enum sanitize the value.
        If obj is a primitive return directly.
        If obj is a date or datetime convert to string in iso8601 format.
        If obj is a UUID convert to string.
        If obj is a list or tuple, sanitize each element.
        If obj is a dict, sanitize the dict.
        If obj is an API model, convert to dict.

        :param obj: The data to serialize.

        :return: The serialized form of data.
        """
        if obj is None:
            # If obj is None return None.
            return None
        if isinstance(obj, Enum):
            # If obj is an Enum sanitize the value.
            return cls._sanitize_for_serialization(obj.value)
        elif isinstance(obj, (str, int, float, bool, bytes)):
            # If obj is a primitive return directly.
            return obj
        elif isinstance(obj, (datetime.date, datetime.datetime)):
            # If obj is a date or datetime convert to string in iso8601 format.
            return obj.isoformat()
        elif isinstance(obj, UUID):
            # If obj is a UUID convert to string.
            return str(obj)
        elif isinstance(obj, list):
            # If obj is a list sanitize each element in the list.
            return [cls._sanitize_for_serialization(sub_obj) for sub_obj in obj]
        elif isinstance(obj, tuple):
            # If obj is a tuple sanitize each element in the tuple.
            return tuple(cls._sanitize_for_serialization(sub_obj) for sub_obj in obj)

        if isinstance(obj, Mapping):
            # If obj is a dict, sanitize the dict.
            obj_dict = obj
        elif isinstance(obj, BaseModel):
            # If obj is an API model, convert to dict.
            obj_dict = obj.model_dump(mode="json", by_alias=True, exclude_unset=True)
        else:
            raise ClientTypeError(
                msg=f"{type(obj)} could not be serialized.",
                valid_classes=(
                    NoneType,
                    str,
                    int,
                    float,
                    bool,
                    bytes,
                    list,
                    tuple,
                    dict,
                    BaseModel,
                ),
            )

        return {str(key): cls._sanitize_for_serialization(val) for key, val in obj_dict.items()}

    @classmethod
    def _parameters_to_tuples(
        cls, params: Mapping | list[tuple[str, Any]], collection_formats: Mapping[str, str]
    ) -> list[tuple[str, Any]]:
        """Get parameters as list of tuples, formatting collections.

        :param params: Parameters as dict or list of two-tuples.
        :param collection_formats: Parameter collection formats. One of 'multi', 'ssv', 'pipes', or 'csv'.
            'csv' is the default collection format.

        :return: Parameters as list of tuples, collections formatted
        """
        params = cls._sanitize_for_serialization(params)
        if not isinstance(collection_formats, dict):
            raise ClientTypeError("collection_formats must be a dict.", valid_classes=(dict,))
        new_params = []
        for key, value in list(params.items() if isinstance(params, dict) else params):
            if isinstance(value, (list, tuple)):
                # Implements OpenAPI multi-value parameter styles. Defaults to simple (csv).
                # https://swagger.io/specification/#style-values
                collection_format = collection_formats.get(key, "csv")
                if collection_format == "multi":
                    new_params.extend((key, value) for value in value)
                else:
                    if collection_format == "ssv":
                        delimiter = " "
                    elif collection_format == "pipes":
                        delimiter = "|"
                    else:  # csv is the default.
                        delimiter = ","
                    new_params.append((key, delimiter.join(value)))
            else:
                new_params.append((key, value))
        return new_params

    @classmethod
    def _parameters_to_url_query(cls, params: dict | list[tuple[str, Any]], collection_formats: dict[str, str]) -> str:
        """Get parameters as list of tuples, formatting collections.

        :param params: Parameters as dict or list of two-tuples.
        :param collection_formats: Parameter collection formats. One of 'multi', 'ssv', tsv', 'pipes', or 'csv'.
            'csv' is the default collection format.

        :return: URL query string (e.g. a=Hello%20World&b=123)
        """
        query_params = cls._parameters_to_tuples(params, collection_formats)
        return urlencode(query_params)

    @classmethod
    def _deserialize(cls, response: HTTPResponse, response_type: type[T] | None) -> T:
        """Deserializes dict, list, or str into an object.

        :param response: Value as dict, list, or str.
        :param response_type: Target type to deserialize data to. Can be class literal, list[T], or dict[str, T].

        :return: The deserialized object.
        """
        if response_type is not None and isclass(response_type) and issubclass(response_type, HTTPResponse):
            # Return the response object directly.
            return response

        if response_type is not bytes:
            match = None
            content_type = response.getheader("content-type")
            if content_type is not None:
                match = re.search(r"charset=([a-zA-Z\-\d]+)[\s;]?", content_type)
            encoding = match.group(1) if match else "utf-8"
            response_data = response.data.decode(encoding)
        else:
            response_data = response.data

        try:
            response_data = json.loads(response_data)
        except ValueError:
            pass  # data must not be JSON formatted.

        if response_type is None:  # Return decoded data for a known response without a schema.
            return response_data
        elif response_type is EmptyResponse and response_data == "":
            return EmptyResponse(status=response.status, reason=response.reason, headers=response.getheaders())
        elif response_type is EmptyResponse and response_data != "":
            raise ClientValueError(msg=f"Unexpected content with '{response.status}' status code")
        elif isclass(response_type) and issubclass(response_type, EvoAPIException):  # Raise if API exception.
            if response_type is EvoAPIException and CustomTypedError.provided_by(response_data):
                # Use Typed Error types, but don't override generalized error types.
                response_type = CustomTypedError.from_type_id(response_data.get("type"))
            raise response_type(
                status=response.status, reason=response.reason, content=response_data, headers=response.headers
            )
        else:
            try:
                return cls.__deserialize(response_data, response_type)
            except EvoClientException:
                raise
            except Exception as e:
                raise ClientValueError(msg="Could not deserialize result", caused_by=e)

    @classmethod
    def __deserialize(cls, data: dict | list | str, response_type: type[T]) -> T:
        """Deserializes dict, list, or str into an object.

        :param data: Value as dict, list, or str.
        :param response_type: Target type to deserialize data to. Can be class literal, list[T], or dict[str, T].

        :return: The deserialized object.
        """
        if data is None:
            return None

        if isinstance(response_type, GenericAlias):  # list[T], dict[str, T].
            return cls.__deserialize_generic(data, response_type)
        elif response_type in {str, int, float, bool, bytes, dict}:
            return cls.__deserialize_primitive(data, response_type)
        elif response_type is datetime.datetime:
            return cls.__deserialize_datetime(data)
        elif response_type is datetime.date:
            return cls.__deserialize_datetime(data).date()
        elif response_type is datetime.time:
            return cls.__deserialize_time(data)
        elif issubclass(response_type, BaseModel):  # API Models.
            return response_type.model_validate(data)
        else:
            raise ClientValueError("Could not parse content.")

    @classmethod
    def __deserialize_generic(cls, data: list | dict, klass: GenericAlias) -> list | dict[str, Any]:
        """Deserializes list or dict into a list or dict of objects.

        :param data: Value as list or dict.
        :param klass: Target type to deserialize data to. Can be list[T] or dict[str, T].

        :return: The deserialized object.

        :raises ClientTypeError: If the data could not be deserialized.
        """
        if klass.__origin__ is list and isinstance(data, list):
            # Deserialize list.
            (inner_klass,) = klass.__args__
            return [cls.__deserialize(sub_data, inner_klass) for sub_data in data]

        elif klass.__origin__ is dict and klass.__args__[0] is str and isinstance(data, dict):
            # Deserialize dict.
            _, value_klass = klass.__args__
            return {str(key): cls.__deserialize(value, value_klass) for key, value in data.items()}

        else:
            raise ClientTypeError(msg=f"Could not deserialize '{type(data)}' as '{klass}'.")

    @staticmethod
    def __deserialize_primitive(data: str, klass: type):
        """Deserializes string to primitive type.

        :param data: str.
        :param klass: class literal.

        :return: str, int, float, bool.

        :raises ClientTypeError: If the data could not be deserialized.
        """
        try:
            return klass(data)
        except TypeError as e:
            raise ClientTypeError(msg="Could not deserialize primitive", caused_by=e)

    @staticmethod
    def __deserialize_datetime(string: str) -> datetime.datetime:
        """Deserializes string to datetime.

        :param string: Date string.

        :return: datetime object.

        :raises ClientValueError: If the string could not be parsed as datetime.
        """
        try:
            return parse(string)
        except ValueError as e:
            raise ClientTypeError(msg="Could not deserialize datetime", caused_by=e)

    @classmethod
    def __deserialize_time(cls, string: str) -> datetime.time:
        """Deserializes string to time.

        :param string: time string.

        :return: time object.

        :raises ClientValueError: If the string could not be parsed as datetime.
        """
        datetime_value = cls.__deserialize_datetime(string)
        return datetime_value.time().replace(tzinfo=datetime_value.tzinfo)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import copy
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias

if TYPE_CHECKING:
    from typing import Self

    from .data import HTTPHeaderDict


class EvoClientException(Exception):
    """The base exception class for all Evo client exceptions."""


_Condition: TypeAlias = type[Exception] | tuple[type[Exception], ...] | Callable[[Exception], bool]


def _get_filter_func(condition: _Condition) -> Callable[[Exception], bool]:
    if not isinstance(condition, type) and callable(condition):
        return condition
    else:
        return lambda e: isinstance(e, condition)


class EvoExceptionGroup(EvoClientException):
    """A custom exception group type that closely resembles `ExceptionGroup` which was introduced in python 3.11

    This implementation is based off PEP 654 and the reference implementation linked within.
    https://github.com/iritkatriel/cpython/pull/10/files#diff-96c4ad902cdfac48712fef8fff0d0c78a71eb135fb48264f3ebb67db1013fc94

    EvoExceptionGroup wraps the exceptions in the sequence excs. The msg parameter must be a string.
    """

    def __new__(cls, msg: str, excs: Sequence[Exception]):
        grp = super().__new__(cls)
        grp._msg = msg
        grp._excs = tuple(excs)
        return grp

    @property
    def message(self) -> str:
        """The msg argument to the constructor. This is a read-only attribute."""
        return self._msg

    @property
    def exceptions(self) -> tuple[Exception, ...]:
        """A tuple of the exceptions in the excs sequence given to the constructor. This is a read-only attribute."""
        return copy.copy(self._excs)

    def derive(self, excs: Sequence[Exception]) -> Self:
        """Create a new instance of the current exception group type with the same message, but wrapping the exceptions
        in excs.

        This method is used by subgroup() and split(). A subclass needs to override it in order to make subgroup() and
        split() return instances of the subclass rather than ExceptionGroup.

        subgroup() and split() copy the __traceback__, __cause__, and __context__ fields from the original
        exception group to the one returned by derive(), so these fields do not need to be updated by derive().

        :param excs: Exceptions to wrap in the derived exception group.

        :return: An exception group with the same message, but which wraps the exceptions in excs.
        """
        return EvoExceptionGroup(self.message, excs)

    def _derive_or_none(self, excs: Sequence[Exception]) -> Self | None:
        """Derive a new exception group with the given exceptions, or return None.

        :param excs: Exceptions to include in the derived exception group.

        :return: A new instance of the current type, with __traceback__, __cause__, and __context__ fields copied. If
            excs is empty, None is returned instead.
        """
        if len(excs) == 0:
            return None
        else:
            derived = self.derive(excs)
            # Copy the original context, cause, and traceback.
            derived.__context__ = copy.copy(self.__context__)
            derived.__cause__ = copy.copy(self.__cause__)
            derived.__traceback__ = copy.copy(self.__traceback__)
            return derived

    def split(self, condition: _Condition) -> tuple[Self | None, Self | None]:
        """Like subgroup(), but returns the pair (match, rest) where match is subgroup(condition) and rest is the
        remaining non-matching part.

        :param condition: Either a function that accepts an exception and returns true for those that should be in
            the subgroup, or it can be an exception type or a tuple of exception types, which is used to check for a
            match using the same check that is used in an except clause.

        :return: (match, rest) where match is subgroup(condition) and rest is the remaining non-matching part.
        """
        filter_func = _get_filter_func(condition)
        if filter_func(self):
            return self._derive_or_none(self.exceptions), None
        matched = []
        unmatched = []
        for exc in self.exceptions:
            if filter_func(exc):
                # Satisfies condition.
                matched.append(exc)
            elif isinstance(exc, EvoExceptionGroup):
                # Recursive splitting.
                matched_subgroup, unmatched_subgroup = exc.split(filter_func)
                if matched_subgroup is not None:
                    matched.append(matched_subgroup)
                if unmatched_subgroup is not None:
                    unmatched.append(unmatched_subgroup)
            else:
                # Does not satisfy condition.
                unmatched.append(exc)

        return self._derive_or_none(matched), self._derive_or_none(unmatched)

    def subgroup(self, condition: _Condition) -> Self | None:
        """Returns an exception group that contains only the exceptions from the current group that match condition, or
        None if the result is empty.

        The nesting structure of the current exception is preserved in the result, as are the values of its message,
        __traceback__, __cause__, and __context__ fields. Empty nested groups are omitted from the result.

        The condition is checked for all exceptions in the nested exception group, including the top-level and any
        nested exception groups. If the condition is true for such an exception group, it is included in the result in
        full.

        :param condition: Either a function that accepts an exception and returns true for those that should be in
            the subgroup, or it can be an exception type or a tuple of exception types, which is used to check for a
            match using the same check that is used in an except clause.

        :return: An exception group that contains only the exceptions from the current group that match condition, or
            None if the result is empty.
        """
        return self.split(condition)[0]

    def __str__(self) -> str:
        excs = self.exceptions
        n_sub_excs = len(excs)
        tb_lines = [
            f"{self.__class__.__name__}: {self.message} ({n_sub_excs} sub-exception{'' if n_sub_excs == 1 else 's'})"
        ]
        for i, exc in enumerate(excs):
            tb_lines.append(f"+---------------- {i + 1} ----------------")
            tb_lines.append(f"| {type(exc).__name__}:")
            for exc_line in str(exc).split("\n"):
                tb_lines.append(f"| {exc_line}")
        return "\n".join(tb_lines)


class RetryError(EvoExceptionGroup):
    """Custom ExceptionGroup for wrapping exceptions from multiple retry attempts."""


class StorageFileNotFoundError(FileNotFoundError, EvoClientException):
    """Raised when a file or directory is requested but doesn't exist."""


class StorageFileExistsError(FileExistsError, EvoClientException):
    """Raised when trying to create a file or directory which already exists."""


class FileNameTooLongError(EvoClientException):
    """Raised when a user tries to download a file with a name that is too long to be saved in the filesystem"""


class ServiceHealthCheckFailed(EvoClientException):
    """Raised when a service health check fails."""


class _WrappedError(EvoClientException):
    """Wrapper for standard exceptions that occur while parsing service responses."""

    def __init__(self, msg: str, caused_by: Exception | None = None):
        """
        :param msg: The exception message.
        :param caused_by: The original error that caused deserialization to fail.
        """
        self.caused_by = caused_by
        full_msg = msg
        if caused_by:
            full_msg = f"{msg}: {str(caused_by)}"
        super(_WrappedError, self).__init__(full_msg)


class TransportError(_WrappedError):
    """Wraps errors raised by the underlying HTTP transport."""


class ClientTypeError(_WrappedError, TypeError):
    """Raised when an operation or function is applied to an object of inappropriate type.
    The associated value is a string giving details about the type mismatch.
    """

    def __init__(
        self,
        msg: str,
        caused_by: Exception | None = None,
        valid_classes: tuple[type, ...] | None = None,
        key_type: bool | None = None,
    ):
        """
        :param msg: The exception message.
        :param caused_by: The original error that caused deserialization to fail.
        :param valid_classes: The primitive classes that current item should be an instance of.
        :param key_type: False if our value is a value in a dict. True if it is a key in a dict. False if our item is an
            item in a list.
        """
        super(ClientTypeError, self).__init__(msg, caused_by)
        self.valid_classes = valid_classes
        self.key_type = key_type


class ClientValueError(_WrappedError, ValueError):
    """Raised when an operation or function receives an argument that has the right type but an inappropriate value,
    and the situation is not described by a more precise exception such as IndexError.
    """


class EvoAPIException(EvoClientException):
    """Base class for all service errors."""

    def __init__(self, status: int, reason: str | None, content: object | None, headers: HTTPHeaderDict | None):
        """
        :param status: HTTP status code.
        :param reason: Reason.
        :param content: Deserialized content from the response.
        :param headers: Response headers
        """
        self.status = status
        self.reason = reason
        self.content = content
        self.headers = headers

    def __str__(self) -> str:
        error_message = f"({self.status})"
        if reason := self.reason:
            error_message += f" {reason}"
        if content := self.content:
            error_message += f"\n{content}"
        return error_message


class UnknownResponseError(EvoAPIException):
    """The service sent an unknown response."""


class BaseTypedError(EvoAPIException):
    """Base class for service errors that have a defined type.

    Concrete error types must subclass BaseTypedError and define the class attribute `TYPE_ID`, which will be used
    to map service error responses to the corresponding implementation.

    The fallback type `DefaultTypedError` will be used if an error response has a title and no other concrete type
    matches the type ID in the response. It is worth noting that the type may be missing from the response,
    in which case `DefaultTypedError` will be used, and the instance `type_` will be 'about:blank' (as is specified).
    """

    status: int
    """The HTTP status code generated by the origin server for this occurrence of the problem."""

    content: dict[str, Any]
    """Deserialized content from the response."""

    @property
    def type_(self) -> str:
        """A URI reference that identifies the problem type."""
        return str(self.content["type"])

    @property
    def title(self) -> str:
        """A short, human-readable summary of the problem type.

        This is advisory only.
        """
        return self.content["title"]

    @property
    def detail(self) -> str | None:
        """A human-readable explanation specific to this occurrence of the problem."""
        return self.content.get("detail")

    def __str__(self) -> str:
        error_message = f"Error: ({self.status})"
        if reason := self.reason:
            error_message += f" {reason}"
        error_message += f"\nType: {self.type_}"
        error_message += f"\nTitle: {self.title}"
        if detail := self.detail:
            error_message += f"\nDetail: {detail}"
        return error_message


class CustomTypedError(BaseTypedError):
    """Base class for service errors that have a type.

    Concrete error types must subclass CustomTypedError and define the class attribute `TYPE_ID`, which will be used
    to map service error responses to the corresponding implementation.

    The fallback type `DefaultTypedError` will be used if an error response has a title and no other concrete type
    matches the type ID in the response. It is worth noting that the type may be missing from the response,
    in which case `DefaultTypedError` will be used, and the instance `type_` will be 'about:blank' (as is specified).
    """

    __CONCRETE_TYPES: dict[str, type[CustomTypedError]] = {}

    TYPE_ID: ClassVar[str]
    """The type ID used to match concrete error types.

    Subclasses should omit the URL portion of the type ID, in case the URL is changed in the future. The full type ID,
    as provided in the response, will be used by the `type_` property.
    """

    @staticmethod
    def from_type_id(type_id: str | None) -> type[CustomTypedError]:
        """Get a concrete error type, based on the type ID.

        :param type_id: The type ID of the received error.

        :return: The concrete implementation of the requested type.
        """
        if type_id is None:
            return DefaultTypedError

        for defined_type_id, error_type in CustomTypedError.__CONCRETE_TYPES.items():
            if type_id.endswith(defined_type_id):
                return error_type
        return DefaultTypedError

    @staticmethod
    def provided_by(content: object) -> bool:
        """Determine whether the content of an error response has a title

        :param content: The deserialized content of a 4xx or 5xx error response.

        :return: True if the content satisfies has a title.
        """
        return isinstance(content, dict) and "title" in content

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        try:
            type_id = cls.TYPE_ID
        except AttributeError:
            raise ValueError(f"{cls} must define TYPE_ID.")
        if type_id is None:
            return  # abstract class
        if not isinstance(type_id, str):
            raise ValueError(f"{cls} TYPE_ID must be a string.")
        if existing_cls := CustomTypedError.__CONCRETE_TYPES.get(type_id):
            raise ValueError(f"Duplicated TYPE_ID between {cls} and {existing_cls}")
        CustomTypedError.__CONCRETE_TYPES[type_id] = cls


class DefaultTypedError(CustomTypedError):
    """Fallback Typed Error implementation that meets the minimum specification."""

    TYPE_ID = "about:blank"

    @property
    def type_(self) -> str:
        """A URI reference that identifies the problem type.

        When this member is not present, its value is assumed to be `about:blank`. This is the primary identifier for
        the problem type.
        """
        return self.content.get("type", self.TYPE_ID)


class GeneralizedTypedError(BaseTypedError):
    """Base class for Typed errors that should be generalized based on status code.

    Generalized error types must subclass GeneralizedTypedError and define the class attribute `STATUS_CODE`, which will
    be used to map service error codes to the corresponding generalization.

    This mechanism allows specific error codes to be treated consistently, regardless of service implementation.
    """

    __GENERALIZED_TYPES: dict[int, type[GeneralizedTypedError]] = {}

    TYPE_ID = "about:blank"
    STATUS_CODE: int

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        status_code = getattr(cls, "STATUS_CODE")
        GeneralizedTypedError.__GENERALIZED_TYPES[status_code] = cls

    @staticmethod
    def from_status_code(status_code: int) -> type[GeneralizedTypedError] | None:
        """Get a generalized error type, based on the status code.

        :param status_code: The status code of the error response.

        :return: The generalized implementation for the requested status code.
        """
        return GeneralizedTypedError.__GENERALIZED_TYPES.get(status_code, None)

    @property
    def type_(self) -> str:
        """A URI reference that identifies the problem type.

        When this member is not present, its value is assumed to be `about:blank`. This is the primary identifier for
        the problem type.
        """
        if isinstance(self.content, dict) and "type" in self.content:
            return self.content["type"]
        else:
            return self.TYPE_ID

    @property
    def title(self) -> str:
        """A short, human-readable summary of the problem type.

        This is advisory only.
        """
        if isinstance(self.content, dict) and "title" in self.content:
            return str(self.content["title"])
        else:
            return self.reason

    @property
    def detail(self) -> str | None:
        """A human-readable explanation specific to this occurrence of the problem."""
        if isinstance(self.content, dict) and "detail" in self.content:
            return str(self.content["detail"])
        else:
            return None


class BadRequestException(GeneralizedTypedError):
    """The service cannot process the request due to a client error (400 - Bad Request)."""

    STATUS_CODE = 400


class UnauthorizedException(GeneralizedTypedError):
    """The client must authenticate to get a response (401 - Unauthorized)."""

    STATUS_CODE = 401


class ForbiddenException(GeneralizedTypedError):
    """The client does not have access rights to the content (403- Forbidden)."""

    STATUS_CODE = 403


class NotFoundException(GeneralizedTypedError):
    """The Service API could not find the requested resource (404 - Not Found)."""

    STATUS_CODE = 404


class GoneException(GeneralizedTypedError):
    """The requested resource is deleted (410 - Gone)."""

    STATUS_CODE = 410


class SelectionError(EvoClientException):
    """Raised when a selection error occurs."""



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from typing import Any
from unittest import mock
from urllib.parse import urljoin

from ..connector import APIConnector
from ..data import HTTPHeaderDict, HTTPResponse, RequestMethod
from ..interfaces import IAuthorizer, ITransport
from .consts import ACCESS_TOKEN, BASE_URL


class TestHTTPHeaderDict(HTTPHeaderDict):
    """A subclass of HTTPHeaderDict that overrides the __repr__ method to provide a more readable output in unit tests.

    The sole benefit of this class is that it does not hide sensitive information when printed, making it easier to
    visually compare the actual and expected values in unit tests.
    """

    __test__ = False  # Prevent unittest from discovering this class as a test case.

    def __repr__(self) -> str:
        repr_data = {key.title(): value for key, value in self.items()}
        return f"{self.__class__.__name__}({repr_data!r})"


class MockResponse(mock.Mock):
    """Fake HTTPResponse object to be returned by a mocked IHttpTransport"""

    def __init__(
        self,
        status_code: int,
        reason: str | None = None,
        headers: Mapping[str, str] | None = None,
        body: bytes | None = None,
        content: str = "",
    ):
        """
        :param status_code: HTTP status code.
        :param reason: Response reason.
        :param headers: Response headers.
        :param body: Response body, as bytes. Body takes precedence over content.
        :param content: Response body, as a string. Content will be encoded using utf-8.
        """
        super().__init__(spec=HTTPResponse)
        self.status = status_code
        self._content = content
        self.headers = TestHTTPHeaderDict(headers)
        self.data = body if body is not None else content.encode("utf-8")
        self.reason = reason
        self.getheader = mock.Mock(side_effect=self._get_header)
        self.getheaders = mock.Mock(return_value=self.headers.copy())

    def _get_header(self, name: str, default: Any | None = None) -> Any:
        return self.headers.get(name, default)


class AbstractTestRequestHandler(ABC):
    """Base class for a test request handler that can be attached to TestTransport."""

    @abstractmethod
    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        """Same as ITransport.request(), but returns a MockResponse object.

        TestTransport attaches this handler to the request method, so that it can be used to simulate complex API
        behaviour in tests.

        :param method: HTTP request method.
        :param url: HTTP request url.
        :param headers: Http request headers.
        :param body: Request body.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The HTTP response object.

        :raise ClientValueError: If `body` and `post_params` are both provided.
        :raise TransportError: If the underlying implementation encounters an error.
        """
        ...  # pragma: no cover

    @staticmethod
    def not_found() -> MockResponse:
        return MockResponse(status_code=404, reason="Not Found")

    @staticmethod
    def bad_request() -> MockResponse:
        return MockResponse(status_code=400, reason="Bad Request")


class TestTransport(mock.AsyncMock):
    """Fake ITransport object to be used in tests"""

    open: mock.AsyncMock
    close: mock.AsyncMock
    request: mock.AsyncMock

    def __init__(self, *, base_url: str = BASE_URL) -> None:
        super().__init__(spec=ITransport)
        self._base_url = base_url
        self.request.return_value = MockResponse(status_code=503)

    def _join_hostname(self, path: str) -> str:
        return urljoin(self._base_url, path.lstrip("/"))

    @contextmanager
    def set_http_response(
        self,
        status_code: int,
        content: str = "",
        reason: str | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> Iterator[MockResponse]:
        """Context manager constructor to temporarily use a different MockResponse with the provided parameters.

        :param status_code: HTTP status code.
        :param content: Response content, as a string. Content will be encoded using utf-8.
        :param reason: Response reason.
        :param headers: Response headers.

        :yields: the constructed MockResponse object.
        """
        old_response = self.request.return_value
        self.request.return_value = new_response = MockResponse(
            status_code=status_code, content=content, reason=reason, headers=headers
        )
        old_side_effect = self.request.side_effect
        self.request.side_effect = None
        yield new_response
        self.request.side_effect = old_side_effect
        self.request.return_value = old_response

    def set_request_handler(self, handler: AbstractTestRequestHandler) -> None:
        """Set the request handler for this transport.

        :param handler: The request handler to be used for all requests.
        """
        self.request.side_effect = handler.request

    def assert_request_made(
        self,
        method: RequestMethod,
        path: str = "",
        headers: Mapping[str, str] | None = None,
        post_params: list[tuple[str, str]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> None:
        """Assert that last call to ITransport.request() used the specified arguments.

        :param method: HTTP request method.
        :param path: The API path, relative to the base_url.
        :param headers: Http request headers.
        :param body: Request json body, for `application/json`.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.
        """
        headers = TestHTTPHeaderDict(headers or {})
        self.request.assert_called_with(
            method=method,
            url=self._join_hostname(path),
            headers=headers,
            post_params=post_params,
            body=body,
            request_timeout=request_timeout,
        )

    def assert_any_request_made(
        self,
        method: RequestMethod,
        path: str = "",
        headers: Mapping[str, str] | None = None,
        post_params: list[tuple[str, str]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> None:
        """Assert that any call to ITransport.request() used the specified arguments.

        :param method: HTTP request method.
        :param path: The API path, relative to the base_url.
        :param headers: Http request headers.
        :param body: Request json body, for `application/json`.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.
        """
        headers = TestHTTPHeaderDict(headers or {})
        self.request.assert_any_call(
            method=method,
            url=self._join_hostname(path),
            headers=headers,
            post_params=post_params,
            body=body,
            request_timeout=request_timeout,
        )

    def assert_n_requests_made(self, n: int) -> None:
        """Assert that there were a certain number of calls to ITransport.request().

        :param n: Number of expected calls.
        """
        assert self.request.await_count == n

    def assert_no_requests(self) -> None:
        """Assert that ITransport.request() has not been called."""
        self.request.assert_not_called()


class TestAuthorizer(mock.AsyncMock):
    """Fake IAuthorizer object to be used in tests"""

    def __init__(self) -> None:
        super().__init__(spec=IAuthorizer)
        self.default_headers = TestHTTPHeaderDict({"Authorization": f"Bearer {ACCESS_TOKEN}"})
        self.refresh_token.return_value = False
        self.get_default_headers.side_effect = lambda: self.default_headers.copy()

    def set_next_access_token(self, token: str) -> None:
        def refresh_token() -> bool:
            self.default_headers = TestHTTPHeaderDict({"Authorization": f"Bearer {token}"})
            self.refresh_token.side_effect = None
            return True

        self.refresh_token.side_effect = refresh_token


class TestWithConnector(unittest.IsolatedAsyncioTestCase):
    """Base unittest class for testing API calls"""

    def setUp(self) -> None:
        self.transport = TestTransport()
        self.authorizer = TestAuthorizer()
        self.connector = APIConnector(BASE_URL, self.transport, self.authorizer)
        self.universal_headers = TestHTTPHeaderDict({})

    def setup_universal_headers(self, headers: Mapping[str, str]) -> None:
        """Set universal headers that are expected to be sent with every request.

        This method should only be called during test setup (i.e., in setUp method).

        :param headers: The headers to be added to every request.
        """
        self.universal_headers = TestHTTPHeaderDict(headers)

    def _get_expected_headers(self, headers: Mapping[str, str] | None) -> TestHTTPHeaderDict:
        return TestHTTPHeaderDict(
            {
                **self.authorizer.default_headers,
                **self.universal_headers,
                **(self.connector._additional_headers or {}),
                **(headers or {}),
            }
        )

    def assert_request_made(
        self,
        method: RequestMethod,
        path: str = "",
        headers: Mapping[str, str] | None = None,
        post_params: list[tuple[str, str]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> None:
        """Assert that last call to ITransport.request() used the specified arguments.

        :param method: HTTP request method.
        :param path: The API path, relative to the base_url.
        :param headers: Http request headers.
        :param body: Request json body, for `application/json`.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.
        """
        self.transport.assert_request_made(
            method=method,
            path=path,
            headers=self._get_expected_headers(headers),
            post_params=post_params,
            body=body,
            request_timeout=request_timeout,
        )

    def assert_any_request_made(
        self,
        method: RequestMethod,
        path: str = "",
        headers: Mapping[str, str] | None = None,
        post_params: list[tuple[str, str]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> None:
        """Assert that any call to ITransport.request() used the specified arguments.

        :param method: HTTP request method.
        :param path: The API path, relative to the base_url.
        :param headers: Http request headers.
        :param body: Request json body, for `application/json`.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.
        """
        self.transport.assert_any_request_made(
            method=method,
            path=path,
            headers=self._get_expected_headers(headers),
            post_params=post_params,
            body=body,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .client import WorkspaceAPIClient
from .data import (
    BasicWorkspace,
    BoundingBox,
    Coordinate,
    OrderByOperatorEnum,
    ServiceUser,
    User,
    UserPermission,
    UserRole,
    Workspace,
    WorkspaceOrderByEnum,
    WorkspaceRole,
)

__all__ = [
    "BasicWorkspace",
    "BoundingBox",
    "Coordinate",
    "OrderByOperatorEnum",
    "ServiceUser",
    "User",
    "UserPermission",
    "UserRole",
    "Workspace",
    "WorkspaceAPIClient",
    "WorkspaceOrderByEnum",
    "WorkspaceRole",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

import dotenv

from evo import logging
from evo.common import ICache

__all__ = ["DotEnv"]

logger = logging.getLogger(__name__)

_KEY = re.compile(r"^[A-Z0-9_\-\.]+$", flags=re.IGNORECASE)


class DotEnv:
    """A simple wrapper around `python-dotenv` to manage and cache environment variables."""

    def __init__(self, cache: ICache, filename: str = ".env") -> None:
        self._file = cache.root / filename
        self._file.touch(exist_ok=True)
        self._cache = dotenv.dotenv_values(self._file, encoding="utf-8")

    def get(self, key: str, default: str | None = None) -> str | None:
        logger.debug(f"Getting environment variable: {key}")
        return self._cache.get(key, default)

    def set(self, key: str, value: str | None) -> None:
        logger.debug(f"Updating environment variable: {key}={value!r}")
        if not _KEY.match(key):
            raise ValueError(f"Invalid key: {key!r}")
        elif value is not None and not isinstance(value, str):
            raise ValueError(f"Invalid value for key {key}: {value!r}")
        elif value is None:
            dotenv.unset_key(self._file, key)
            self._cache.pop(key, None)
        else:
            dotenv.set_key(self._file, key, value, quote_mode="always", encoding="utf-8")
            self._cache[key] = value



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Workspaces API
=============

The Workspaces API enables users to organize, maintain, and store project data, but does not use or process the data. The workspace APIs allow you to manage:
- Workspaces
- User roles within workspaces
- Workspace thumbnails

There are three pre-defined roles within workspaces:

- Owner: can perform all actions in the workspace
- Editor: can perform all actions excluding deleting of a workspace
- Viewer: can view the workspace

These user roles can be assigned to users in a workspace. Once a role has been assigned it can be replaced or removed.
Users can also retrieve user roles, the role of a particular user, and their own role if applicable.
For more information on using the Workspaces API, see the [Workspaces API overview](https://developer.seequent.com/docs/guides/workspaces/), or the API references here.


This code is generated from the OpenAPI specification for Workspaces API.
API version: 1.0
"""

from evo.common.connector import APIConnector
from evo.common.data import RequestMethod
from evo.common.utils import get_header_metadata

from ..models import *  # noqa: F403

__all__ = ["GeneralApi"]


class GeneralApi:
    """API client for the General endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def health_check_health_check_get(
        self,
        full: bool | None = None,
        check_dependencies: bool | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> dict:
        """Health check

        Checks if the service is healthy and able to response to requests

        :param full: (optional)
            Example: `False`
        :param check_dependencies: (optional)
            Example: `False`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the query parameters.
        _query_params = {}
        if full is not None:
            _query_params["full"] = full
        if check_dependencies is not None:
            _query_params["check_dependencies"] = check_dependencies

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": dict,
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/health_check",
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def list_coordinate_systems(
        self,
        org_id: str,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> ListCoordinateSystemsResponse:  # noqa: F405
        """List coordinate systems

        Returns a list of coordinate systems for an organization.

        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": ListCoordinateSystemsResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/coordinate_systems",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def openapi_spec(
        self,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> dict:
        """Get OpenAPI schema

        Returns the OpenAPI schema for the Workspace Service API.

        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": dict,
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/openapi.json",
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Workspaces API
=============

The Workspaces API enables users to organize, maintain, and store project data, but does not use or process the data. The workspace APIs allow you to manage:
- Workspaces
- User roles within workspaces
- Workspace thumbnails

There are three pre-defined roles within workspaces:

- Owner: can perform all actions in the workspace
- Editor: can perform all actions excluding deleting of a workspace
- Viewer: can view the workspace

These user roles can be assigned to users in a workspace. Once a role has been assigned it can be replaced or removed.
Users can also retrieve user roles, the role of a particular user, and their own role if applicable.
For more information on using the Workspaces API, see the [Workspaces API overview](https://developer.seequent.com/docs/guides/workspaces/), or the API references here.


This code is generated from the OpenAPI specification for Workspaces API.
API version: 1.0
"""

from evo.common.connector import APIConnector
from evo.common.data import EmptyResponse, RequestMethod
from evo.common.utils import get_header_metadata

from ..models import *  # noqa: F403

__all__ = ["FoldersApi"]


class FoldersApi:
    """API client for the Folders endpoint.

    NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    :param connector: Client for communicating with the API.
    """

    def __init__(self, connector: APIConnector):
        self.connector = connector

    async def create_folder(
        self,
        workspace_id: str,
        org_id: str,
        folder_create_request: FolderCreateRequest,  # noqa: F405
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> FolderResponse:  # noqa: F405
        """Create a folder

        Creates a new folder in the workspace. The folder needs to have a name and exclusively define either a `parent_folder_id` or a `parent_folder_path`.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param folder_create_request:
            Example: `endpoints.FolderCreateRequest()`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "201": FolderResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders",
            path_params=_path_params,
            header_params=_header_params,
            body=folder_create_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def create_folders_in_path(
        self,
        workspace_id: str,
        org_id: str,
        folders_path: str,
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> CreatePathFoldersResponse:  # noqa: F405
        """Ensure the existence of all folders in a path

        Ensures that all folders in the specified path exist. If a folder does not exist, it will be created.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param folders_path: Folders path. This parameter was automatically generated from a wildcard path.
            Example: `'folders_path_example'`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "folders_path": folders_path,
        }

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": CreatePathFoldersResponse,  # noqa: F405
            "201": CreatePathFoldersResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/path/{folders_path}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def delete_folder(
        self,
        folder_id: str,
        workspace_id: str,
        org_id: str,
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> EmptyResponse:
        """Delete a folder

        Deletes the folder with the given ID. This is not a permanent delete, folder can be restored.

        :param folder_id:
            Format: `uuid`
            Example: `'folder_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "folder_id": folder_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {} | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "204": EmptyResponse,
        }

        return await self.connector.call_api(
            method=RequestMethod.DELETE,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/{folder_id}",
            path_params=_path_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_folder_by_id(
        self,
        workspace_id: str,
        folder_id: str,
        org_id: str,
        limit: int | None = None,
        offset: int | None = None,
        deleted: bool | None = None,
        max_depth: int | None = None,
        folder_name: str | None = None,
        created_by: str | None = None,
        created_at: str | None = None,
        updated_by: str | None = None,
        updated_at: str | None = None,
        order_by: str | None = None,
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> GetFolderResponse:  # noqa: F405
        """Get folder by ID

        Returns a folder by its ID. Optionally include its child folders up to a specified depth.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param folder_id:
            Format: `uuid`
            Example: `'folder_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param limit: (optional) Max number of results per page.
            Example: `5000`
        :param offset: (optional) Number of results to skip before returning `limit` number of results.
            Example: `0`
        :param deleted: (optional) Only show folders that have been deleted.
            Example: `False`
        :param max_depth: (optional) Max depth of the folder tree to return with the folder.
            Example: `0`
        :param folder_name: (optional)
            Example: `'folder_name_example'`
        :param created_by: (optional)
            Format: `uuid`
            Example: `'created_by_example'`
        :param created_at: (optional)
            Example: `'created_at_example'`
        :param updated_by: (optional)
            Format: `uuid`
            Example: `'updated_by_example'`
        :param updated_at: (optional)
            Example: `'updated_at_example'`
        :param order_by: (optional) A comma separated list of fields to order by, where the default sort order is ascending. To specify the sort order, prefix the field name with either `asc:` or `desc:` for ascending or descending respectively. Field names can be one of the following known sort fields: `created_at`, `updated_at`, `name`
            Example: `'order_by_example'`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "folder_id": folder_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if limit is not None:
            _query_params["limit"] = limit
        if offset is not None:
            _query_params["offset"] = offset
        if deleted is not None:
            _query_params["deleted"] = deleted
        if max_depth is not None:
            _query_params["max_depth"] = max_depth
        if folder_name is not None:
            _query_params["folder_name"] = folder_name
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if updated_by is not None:
            _query_params["updated_by"] = updated_by
        if updated_at is not None:
            _query_params["updated_at"] = updated_at
        if order_by is not None:
            _query_params["order_by"] = order_by

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": GetFolderResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/{folder_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def get_folder_by_path(
        self,
        workspace_id: str,
        org_id: str,
        folders_path: str,
        limit: int | None = None,
        offset: int | None = None,
        deleted: bool | None = None,
        max_depth: int | None = None,
        folder_name: str | None = None,
        created_by: str | None = None,
        created_at: str | None = None,
        updated_by: str | None = None,
        updated_at: str | None = None,
        order_by: str | None = None,
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> GetFolderResponse:  # noqa: F405
        """Get folder by path.

        Get a folder at the provided path. Optionally include the child folders up to a specified depth.

        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param folders_path: Folders path. This parameter was automatically generated from a wildcard path.
            Example: `'folders_path_example'`
        :param limit: (optional) Max number of results per page.
            Example: `5000`
        :param offset: (optional) Number of results to skip before returning `limit` number of results.
            Example: `0`
        :param deleted: (optional) Only show folders that have been deleted.
            Example: `False`
        :param max_depth: (optional) Max depth of the folder tree to return with the folder.
            Example: `0`
        :param folder_name: (optional)
            Example: `'folder_name_example'`
        :param created_by: (optional)
            Format: `uuid`
            Example: `'created_by_example'`
        :param created_at: (optional)
            Example: `'created_at_example'`
        :param updated_by: (optional)
            Format: `uuid`
            Example: `'updated_by_example'`
        :param updated_at: (optional)
            Example: `'updated_at_example'`
        :param order_by: (optional) A comma separated list of fields to order by, where the default sort order is ascending. To specify the sort order, prefix the field name with either `asc:` or `desc:` for ascending or descending respectively. Field names can be one of the following known sort fields: `created_at`, `updated_at`, `name`
            Example: `'order_by_example'`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "workspace_id": workspace_id,
            "org_id": org_id,
            "folders_path": folders_path,
        }

        # Prepare the query parameters.
        _query_params = {}
        if limit is not None:
            _query_params["limit"] = limit
        if offset is not None:
            _query_params["offset"] = offset
        if deleted is not None:
            _query_params["deleted"] = deleted
        if max_depth is not None:
            _query_params["max_depth"] = max_depth
        if folder_name is not None:
            _query_params["folder_name"] = folder_name
        if created_by is not None:
            _query_params["created_by"] = created_by
        if created_at is not None:
            _query_params["created_at"] = created_at
        if updated_by is not None:
            _query_params["updated_by"] = updated_by
        if updated_at is not None:
            _query_params["updated_at"] = updated_at
        if order_by is not None:
            _query_params["order_by"] = order_by

        # Prepare the header parameters.
        _header_params = {
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": GetFolderResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.GET,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/path/{folders_path}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def move_folder(
        self,
        folder_id: str,
        workspace_id: str,
        org_id: str,
        folder_move_request: FolderMoveRequest,  # noqa: F405
        api_preview: str | None = None,
        preview_api: str | None = None,
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> FolderResponse:  # noqa: F405
        """Move a folder

        Moves a folder to a new location.

        :param folder_id:
            Format: `uuid`
            Example: `'folder_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param folder_move_request:
            Example: `endpoints.FolderMoveRequest()`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "folder_id": folder_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": FolderResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.PATCH,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/{folder_id}",
            path_params=_path_params,
            header_params=_header_params,
            body=folder_move_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )

    async def update_folder(
        self,
        folder_id: str,
        workspace_id: str,
        org_id: str,
        deleted: bool | None = None,
        api_preview: str | None = None,
        preview_api: str | None = None,
        folder_update_request: FolderUpdateRequest | None = None,  # noqa: F405
        additional_headers: dict[str, str] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> FolderResponse:  # noqa: F405
        """Update a folder

        Updates a folders name.

        :param folder_id:
            Format: `uuid`
            Example: `'folder_id_example'`
        :param workspace_id:
            Format: `uuid`
            Example: `'workspace_id_example'`
        :param org_id:
            Format: `uuid`
            Example: `'org_id_example'`
        :param deleted: (optional)
            Example: `True`
        :param api_preview: (optional) Set to \"opt-in\" to be able to use this API.
            Example: `'api_preview_example'`
        :param preview_api: (optional) Set to \"opt-in\" to be able to use this API. This header is being deprecated. Please use the API-Preview header.
            Example: `'preview_api_example'`
        :param folder_update_request: (optional)
            Example: `endpoints.FolderUpdateRequest()`
        :param additional_headers: (optional) Additional headers to send with the request.
        :param request_timeout: (optional) Timeout setting for this request. If one number is provided, it will be the
            total request timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: Returns the result object.

        :raise evo.common.exceptions.BadRequestException: If the server responds with HTTP status 400.
        :raise evo.common.exceptions.UnauthorizedException: If the server responds with HTTP status 401.
        :raise evo.common.exceptions.ForbiddenException: If the server responds with HTTP status 403.
        :raise evo.common.exceptions.NotFoundException: If the server responds with HTTP status 404.
        :raise evo.common.exceptions.BaseTypedError: If the server responds with any other HTTP status between
            400 and 599, and the body of the response contains a descriptive `type` parameter.
        :raise evo.common.exceptions.EvoAPIException: If the server responds with any other HTTP status between 400
            and 599, and the body of the response does not contain a `type` parameter.
        :raise evo.common.exceptions.UnknownResponseError: For other HTTP status codes with no corresponding response
            type in `response_types_map`.
        """
        # Prepare the path parameters.
        _path_params = {
            "folder_id": folder_id,
            "workspace_id": workspace_id,
            "org_id": org_id,
        }

        # Prepare the query parameters.
        _query_params = {}
        if deleted is not None:
            _query_params["deleted"] = deleted

        # Prepare the header parameters.
        _header_params = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        } | get_header_metadata(__name__)
        if api_preview is not None:
            _header_params["API-Preview"] = api_preview
        if preview_api is not None:
            _header_params["Preview-API"] = preview_api
        if additional_headers is not None:
            _header_params.update(additional_headers)

        # Define the collection formats.
        _collection_formats = {}

        _response_types_map = {
            "200": FolderResponse,  # noqa: F405
        }

        return await self.connector.call_api(
            method=RequestMethod.POST,
            resource_path="/workspace/orgs/{org_id}/workspaces/{workspace_id}/folders/{folder_id}",
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=folder_update_request,
            collection_formats=_collection_formats,
            response_types_map=_response_types_map,
            request_timeout=request_timeout,
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from types import TracebackType
from typing import Any, Literal, Mapping, Type, TypeVar

from pydantic import ValidationError

from evo import logging
from evo.common import APIConnector, RequestMethod
from evo.common.exceptions import EvoAPIException
from evo.common.interfaces import ITransport
from evo.common.utils import BackoffIncremental, BackoffMethod, Retry

from .data import AccessToken
from .exceptions import OAuthError

logger = logging.getLogger("oauth")

__all__ = ["OAuthConnector"]

T = TypeVar("T", bound=AccessToken)

DEFAULT_BASE_URI = "https://ims.bentley.com"


class OAuthConnector:
    """OAuth connector, used to authenticate with an OAuth server and obtain or refresh an access token.

    https://developer.bentley.com/apis/overview/authorization/
    """

    def __init__(
        self,
        transport: ITransport,
        client_id: str,
        client_secret: str | None = None,
        base_uri: str = DEFAULT_BASE_URI,
        max_attempts: int = 3,
        backoff_method: BackoffMethod = BackoffIncremental(2),
    ) -> None:
        """
        :param transport: The transport to use for making requests.
        :param client_id: The OAuth client ID, as registered with the OAuth provider.
        :param client_secret: The OAuth client secret, as registered with the OAuth provider. Only used for client credentials grant types.
        :param base_uri: The base URI of the OAuth server, defaults to ims.bentley.com.

        :raises OAuthError: If the issuer URL is invalid.
        """
        self._connector = APIConnector(base_uri, transport)
        self.__client_id = client_id
        self.__client_secret = client_secret
        self._retry = Retry(logger=logger, max_attempts=max_attempts, backoff_method=backoff_method)

    def endpoint(self, endpoint_type: Literal["authorize", "token"]) -> str:
        """
        Returns the relevant OAuth endpoint by endpoint type. Possible values: "authorize", "token".
        """
        match endpoint_type:
            case "authorize":
                return "/connect/authorize"
            case "token":
                return "/connect/token"
        raise OAuthError("Invalid endpoint type provided. Available: authorize, token")

    @property
    def base_uri(self) -> str:
        return self._connector.base_url.rstrip("/")

    @property
    def client_id(self) -> str:
        return self.__client_id

    async def __aenter__(self) -> OAuthConnector:
        await self._connector.open()
        return self

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_value: Exception | None, traceback: TracebackType | None
    ) -> None:
        await self._connector.close()

    async def _call_api(
        self,
        method: RequestMethod,
        resource_path: str,
        path_params: Mapping[str, Any] | None = None,
        query_params: Mapping[str, Any] | None = None,
        header_params: Mapping[str, Any] | None = None,
        post_params: Mapping[str, Any] | None = None,
        body: object | str | bytes | None = None,
        collection_formats: Mapping[str, str] | None = None,
        response_types_map: Mapping[str, type[T]] | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> T:
        """Wrapper for APIConnector.call_api() with retry on 502 and 504 errors."""
        async for attempt in self._retry:
            try:
                return await self._connector.call_api(
                    method,
                    resource_path,
                    path_params=path_params,
                    query_params=query_params,
                    header_params=header_params,
                    post_params=post_params,
                    body=body,
                    collection_formats=collection_formats,
                    response_types_map=response_types_map,
                    request_timeout=request_timeout,
                )
            except EvoAPIException as e:
                if e.status in {502, 504}:
                    logger.warning(f"OAuth {attempt} failed with status {e.status}")
                    attempt.set_exception(e)
                else:
                    raise

    async def fetch_token(self, data: dict, expected_response_model: Type[T]) -> T:
        """Fetch an access token from the server.

        :param data: The data to send to the server.

        :return: The token response.

        :raises OAuthError: If the token response cannot be fetched.
        """
        data["client_id"] = self.client_id
        if self.__client_secret is not None:
            data["client_secret"] = self.__client_secret

        try:
            async with self._connector:
                try:
                    response = await self._call_api(
                        RequestMethod.POST,
                        self.endpoint("token"),
                        header_params={
                            "Accept": "application/json",
                            "Content-Type": "application/x-www-form-urlencoded",
                        },
                        post_params=data,
                        response_types_map={"200": expected_response_model},
                    )
                except EvoAPIException as e:
                    error_json: dict = e.content
                    title = error_json.get("error", "Unexpected response from server")
                    detail = error_json.get("error_description", str(e))
                    raise OAuthError(f"{title}: {detail}")
                except ValidationError as e:
                    raise OAuthError("Invalid token response from server.") from e
            return response
        except OAuthError:
            raise  # Re-raise OAuthError exceptions.
        except Exception as exc:  # noqa: E722
            # Catch any other exceptions and raise OAuthError.
            raise OAuthError("Unable to fetch access token." + str(exc)) from exc



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import os
from collections.abc import Iterator
from dataclasses import dataclass
from threading import Lock

from evo import logging

from ..interfaces import IFeedback
from ..utils import NoFeedback, Retry, RetryHandler
from .exceptions import ChunkedIOError
from .interfaces import IDestination, ISource

logger = logging.getLogger("io.manager")

__all__ = [
    "DEFAULT_CHUNK_SIZE",
    "ChunkMetadata",
    "ChunkedIOManager",
    "ChunkedIOTracker",
]

DEFAULT_CHUNK_SIZE = 4 * 1024 * 1024
"""The default max size of each chunk to be transferred."""


@dataclass(frozen=True)
class ChunkMetadata:
    """Metadata for a chunk of data to be transferred."""

    id: int
    """The chunk ID, starting from 0."""

    offset: int
    """The byte offset where this chunk starts, where 0 is the beginning of the file."""

    size: int
    """The size of this chunk in bytes. The final chunk will usually be smaller than the chunk_size."""

    completed: bool
    """Whether this chunk has been successfully transferred."""


class ChunkedIOTracker:
    """Tracks the progress of chunked data transfer."""

    def __init__(self, file_size: int, chunk_size: int):
        """
        :param file_size: The total size of the file to be transferred.
        :param chunk_size: The maximum size of each chunk to be transferred.
        """
        self._chunk_size = chunk_size
        self._total_size = file_size
        self._n_chunks, rem = divmod(file_size, chunk_size)
        if rem:
            self._n_chunks += 1
        self._n_chunks_completed = 0
        self._chunk_status: list[bool] = [False] * self._n_chunks
        self._lock = Lock()

    def __iter__(self) -> Iterator[ChunkMetadata]:
        """Iterate over the chunks of data to be transferred.

        :return: An iterator of `evo.common.io.ChunkMetadata` objects.
        """
        for i, is_complete in enumerate(self._chunk_status):
            this_offset = i * self._chunk_size
            this_chunk_size = min(self._chunk_size, self._total_size - this_offset)
            yield ChunkMetadata(id=i, offset=this_offset, size=this_chunk_size, completed=is_complete)

    def set_complete(self, chunk: ChunkMetadata) -> None:
        """Mark a chunk as completed (successfully transferred).

        :param chunk: Metadata for the chunk that has been transferred.
        """
        with self._lock:
            self._chunk_status[chunk.id] = True
            self._n_chunks_completed += 1

    def get_progress(self) -> float:
        """Get the percentage of total chunks that have been transferred.

        :return: A float between 0 and 1 representing the progress, rounded to 2 decimal places.
        """
        if self._n_chunks == 0:
            return 1.0
        return round(self._n_chunks_completed / self._n_chunks, 2)

    def is_complete(self) -> bool:
        """Check if all chunks have been successfully transferred.

        :return: True if all chunks have been successfully transferred.
        """
        return self._n_chunks_completed == self._n_chunks


class ChunkedIOManager:
    """Manager for robust multithreaded data transfer from `evo.common.io.interfaces.ISource` to
    `evo.common.io.interfaces.IDestination`.

    This class is designed to transfer large files in chunks, with the ability to retry failed transfers. The progress
    of the transfer can be reported to a `evo.common.interfaces.IFeedback` object.
    """

    def __init__(
        self,
        message: str = "",
        retry: Retry | None = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        max_workers: int | None = None,
    ):
        """
        :param message: The message to update feedback progress with.
        :param retry: A Retry object with a wait strategy. If None, a default Retry is created. If a chunk is
            successfully transferred the attempt counter will be reset.
        :param chunk_size: The maximum size of each chunk to be transferred. The final chunk will usually be smaller
            than the chunk_size.
        :param max_workers: The maximum number of concurrent connections to use for data transfer. If None, the default
            is `min(32, os.cpu_count() + 4)`, which is consistent with `concurrent.futures.ThreadPoolExecutor`.
        """
        self._message = message
        self._retry = Retry(logger=logger) if retry is None else retry
        self._chunk_size = chunk_size
        self._tracker: ChunkedIOTracker | None = None

        if max_workers is None:
            # Same default as concurrent.futures.ThreadPoolExecutor.
            max_workers = min(32, os.cpu_count() + 4)
        self._rate_limiter = asyncio.BoundedSemaphore(max_workers)

    async def _transfer(self, source: ISource, destination: IDestination, chunk: ChunkMetadata) -> float:
        async with self._rate_limiter:
            data = await source.read_chunk(chunk.offset, chunk.size)
            await destination.write_chunk(chunk.offset, data)
            self._tracker.set_complete(chunk)
            return self._tracker.get_progress()

    async def _attempt_transfer(
        self, source: ISource, destination: IDestination, fb: IFeedback, retry_handler: RetryHandler
    ) -> None:
        if self._tracker is None:
            self._tracker = ChunkedIOTracker(await source.get_size(), self._chunk_size)
        last_progress = self._tracker.get_progress()
        fb.progress(last_progress, self._message)

        # Python 3.11 introduces asyncio.TaskGroup, which can be used to make this much cleaner.
        transfers = []
        try:
            # Append rather than list comprehension, because we want to be able to cancel previous tasks if creating a
            # new task fails. With TaskGroup this can become a list comprehension.
            for chunk in self._tracker:
                if chunk.completed:
                    continue
                transfers.append(asyncio.create_task(self._transfer(source, destination, chunk)))

            for transfer in asyncio.as_completed(transfers):
                this_progress = await transfer
                if this_progress != last_progress:
                    fb.progress(this_progress, self._message)
                    last_progress = this_progress
                retry_handler.reset_counter()  # Reset retry counter on successful transfer.
        finally:
            for transfer in transfers:
                if not transfer.done():
                    transfer.cancel()

    async def run(self, source: ISource, destination: IDestination, fb: IFeedback | None = None) -> None:
        """Transfer all data from source to destination.

        Recoverable transfer failures must be raised as a subclass of `evo.common.io.ChunkedIOManager`,
        this will retry recoverable transfers up to the max_attempts specified in Retry after calling the error's
        recover() method. The attempt counter is reset whenever a chunk is successfully transferred.

        If the maximum number of attempts is reached, `evo.common.exceptions.RetryError` will be
        raised with the last error.

        :param source: The source to read data from.
        :param destination: The destination to write data to.
        :param fb: A feedback object for reporting transfer progress.
        """
        fb = NoFeedback if fb is None else IFeedback.adapt(fb, allow_implicit=True)
        async for handler in self._retry:
            with handler.suppress_errors(ChunkedIOError):
                await self._attempt_transfer(source, destination, fb, handler)

            if isinstance(handler.exception, ChunkedIOError):
                fb.progress(0.0, "Retrying...")
                if not await handler.exception.recover():
                    raise handler.exception

    def is_complete(self) -> bool:
        """Check if all chunks have been successfully transferred.

        :return: True if all chunks have been successfully transferred.
        """
        return self._tracker is not None and self._tracker.is_complete()
